# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib.resources
import logging
import math
import numpy as np
import os
from prpy.ffmpeg.probe import probe_video
from prpy.ffmpeg.readwrite import read_video_from_path
from prpy.numpy.image import crop_slice_resize
from typing import Union, Tuple
import urllib.request
import yaml

from vitallens.constants import API_MIN_FRAMES, VIDEO_PARSE_ERROR

def load_config(filename: str) -> dict:
  """Load a yaml config file.

  Args:
    filename: The filename of the yaml config file
  Returns:
    loaded: The contents of the yaml config file
  """
  with importlib.resources.open_binary('vitallens.configs', filename) as f:
    loaded = yaml.load(f, Loader=yaml.Loader)
  return loaded

def download_file(url: str, dest: str):
  """Download a file if necessary.

  Args:
    url: The url to download the file from
    dest: The path to write the downloaded file to
  """
  if not os.path.exists(dest):
    os.makedirs(os.path.dirname(dest), exist_ok=True)
    logging.info("Downloading {} to {}".format(url, dest))
    urllib.request.urlretrieve(url, dest)
  else:
    logging.info("{} already exists, skipping download.".format(dest))

def probe_video_inputs(
    video: Union[np.ndarray, str],
    fps: float = None
  ) -> Tuple[tuple, float]:
  """Check the video inputs and probe to extract metadata.

  Args:
    video: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
      with a sequence of frames in unscaled uint8 RGB format, or a path to a
      video file.
    fps: Sampling frequency of the input video. Required if type(video)==np.ndarray.
  Returns:
    video_shape: The shape of the input video as (n_frames, h, w, 3)
    fps: Sampling frequency of the input video.
    issues: True if a possible issue with the video has been detected.
  """
  # Check that fps is correct type
  if not (fps is None or isinstance(fps, (int, float))):
    raise ValueError("fps should be a number, but got {}".format(type(fps)))
  # Check if video is array or file name
  if isinstance(video, str):
    if os.path.isfile(video):
      try:
        fps_, n, w_, h_, _, _, r, i = probe_video(video)
        if fps is None: fps = fps_
        if abs(r) == 90: h = w_; w = h_
        else: h = h_; w = w_
        return (n, h, w, 3), fps, i
      except Exception as e:
        raise ValueError("Problem probing video at {}: {}".format(video, e))
    else:
      raise ValueError("No file found at {}".format(video))
  elif isinstance(video, np.ndarray):
    if fps is None:
      raise ValueError("fps must be specified for ndarray input")
    if video.dtype != np.uint8:
      raise ValueError("video.dtype should be uint8, but got {}".format(video.dtype))
    if len(video.shape) != 4 or video.shape[0] < API_MIN_FRAMES or video.shape[3] != 3:
      raise ValueError("video should have shape (n_frames [>= {}], h, w, 3), but found {}".format(API_MIN_FRAMES, video.shape))
    return video.shape, fps, False
  else:
    raise ValueError("Invalid video {}, type {}".format(video, type(input)))

def parse_video_inputs(
    video: Union[np.ndarray, str],
    fps: float = None,
    roi: tuple = None,
    target_size: Union[int, tuple] = None,
    target_fps: float = None,
    library: str = 'prpy',
    scale_algorithm: str = 'bilinear',
    trim: tuple = None,
    dim_deltas: tuple = (1, 1, 1)
  ) -> Tuple[np.ndarray, float, tuple, int]:
  """Parse video inputs into required shape.

  Args:
    video: The video input. Either a filepath to video file or ndarray
    fps: Framerate of video input. Can be `None` if video file provided.
    roi: The region of interest as (x0, y0, x1, y1). Use None to keep all.
    target_size: Optional target size as int or tuple (h, w)
    target_fps: Optional target framerate
    library: Library to use for resample if video is np.ndarray
    scale_algorithm: Algorithm to use for resample
    trim: Frame numbers for temporal trimming (start, end) (optional).
    dim_deltas: Maximum acceptable deviation from expected video (n, h, w) dims.
  Returns:
    parsed: Parsed inputs as `np.ndarray` with type uint8. Shape (n, h, w, c)
      if target_size provided, h = target_size[0] and w = target_size[1].
    fps_in: Frame rate of original inputs
    shape_in: Shape of original inputs in form (n, h, w, c)
    ds_factor: Temporal downsampling factor applied
    idxs: The frame indices returned from original video
  """
  # Check if input is array or file name
  if isinstance(video, str):
    if os.path.isfile(video):
      try:
        fps_, n, w_, h_, _, _, r, i = probe_video(video)
        if fps is None: fps = fps_
        if roi is not None: roi = (int(roi[0]), int(roi[1]), int(roi[2]), int(roi[3]))
        if isinstance(target_size, tuple): target_size = (target_size[1], target_size[0])
        if abs(r) == 90: h = w_; w = h_
        else: h = h_; w = w_
        try:
          video, ds_factor = read_video_from_path(
            path=video, target_fps=target_fps, crop=roi, scale=target_size, trim=trim,
            pix_fmt='rgb24', dim_deltas=dim_deltas, scale_algorithm=scale_algorithm)
        except:
          ValueError(VIDEO_PARSE_ERROR)
        expected_n = math.ceil(((trim[1]-trim[0]) if trim is not None else n) / ds_factor)
        if video.shape[0] < expected_n:
          logging.warning("Less frames received than expected (delta = {}) - this may indicate an issue with the video file. Padding to avoid issues.".format(video.shape[0]-expected_n))
          video = np.concatenate((np.repeat(video[:1], expected_n - video.shape[0], axis=0), video), axis=0)
        elif video.shape[0] > expected_n:
          logging.warning("More frames received than expected (delta = {}) - this may indicate an issue with the video file. Trimming to avoid issues.".format(video.shape[0]-expected_n))
          video = video[:expected_n]
        start_idx = max(0, trim[0]) if trim is not None else 0
        end_idx = min(n, trim[1]) if trim is not None else n
        idxs = list(range(start_idx, end_idx, ds_factor))
        if video.shape[0] != expected_n or len(idxs) != expected_n:
          raise ValueError(VIDEO_PARSE_ERROR)
        return video, fps, (n, h, w, 3), ds_factor, idxs
      except Exception as e:
        raise ValueError("Problem reading video from {}: {}".format(video, e))
    else:
      raise ValueError("No file found at {}".format(video))
  elif isinstance(video, np.ndarray):
    video_shape_in = video.shape
    # Downsample / crop / scale if necessary
    ds_factor = 1
    if target_fps is not None:
      if target_fps > fps: logging.warning("target_fps should not be greater than fps. Ignoring.")
      else: ds_factor = max(round(fps / target_fps), 1)
    target_idxs = None if ds_factor == 1 else list(range(video.shape[0])[0::ds_factor])
    if trim is not None:
      if target_idxs is None: target_idxs = range(video_shape_in[0])
      target_idxs = [idx for idx in target_idxs if trim[0] <= idx < trim[1]]
    if roi is not None or target_size is not None or target_idxs is not None:
      if target_size is None and roi is not None: target_size = (int(roi[3]-roi[1]), int(roi[2]-roi[0]))
      elif target_size is None: target_size = (video.shape[1], video.shape[2])
      video = crop_slice_resize(
        inputs=video, target_size=target_size, roi=roi, target_idxs=target_idxs,
        preserve_aspect_ratio=False, library=library, scale_algorithm=scale_algorithm)
    if target_idxs is None: target_idxs = list(range(video_shape_in[0]))
    return video, fps, video_shape_in, ds_factor, target_idxs
  else:
    raise ValueError("Invalid video {}, type {}".format(video, type(video)))

def merge_faces(faces: np.ndarray) -> tuple:
  """Compute the union of all faces.
  
  Args:
    faces: The face detections. Shape (n, 4)
  Returns:
    union: Tuple (x0, y0, x1, y1)
  """
  # Get the minimum x0, y0 and maximum x1, y1 values across all frames
  x0 = np.min(faces[:, 0])
  y0 = np.min(faces[:, 1])
  x1 = np.max(faces[:, 2])
  y1 = np.max(faces[:, 3])
  # Return as tuple
  return (x0, y0, x1, y1)

def check_faces(
    faces: Union[list, np.ndarray],
    inputs_shape: tuple
  ) -> np.ndarray:
  """Make sure the face detections are in a correct format.

  Args:
    faces: The specified faces in form [x0, y0, x1, y1]. Either
      - list/ndarray of shape (n_faces, n_frames, 4) for multiple faces detected on multiple frames
      - list/ndarray of shape (n_frames, 4) for single face detected on mulitple frames
      - list/ndarray of shape (4,) for single face detected globally
      - None to assume frames already cropped to single face
    inputs_shape: The shape of the inputs.
  Returns:
    faces: The faces. np.ndarray of shape (n_faces, n_frames, 4) in form [x_0, y_0, x_1, y_1]
  """
  n_frames, h, w, _ = inputs_shape
  if faces is None:
    # Assume that each entire frame is a single face
    logging.info("No faces given - assuming that frames have been cropped to a single face")
    faces = np.tile(np.asarray([0, 0, w, h], dtype=np.int64), (n_frames, 1))[np.newaxis] # (1, n_frames, 4)
  else:
    faces = np.asarray(faces, dtype=np.int64)
    if faces.shape[-1] != 4: raise ValueError("Face detections must be in flat point form")
    if len(faces.shape) == 1:
      # Single face detection given - repeat for n_frames
      faces = np.tile(faces, (n_frames, 1))[np.newaxis] # (1, n_frames, 4)
    elif len(faces.shape) == 2:
      # Single face detections for multiple frames given
      if faces.shape[0] != n_frames:
        raise ValueError("Assuming detections of a single face for multiple frames given, but number of frames ({}) did not match number of face detections ({})".format(
          n_frames, faces.shape[0]))
      faces = faces[np.newaxis]
    elif len(faces.shape) == 3:
      if faces.shape[1] == 1:
        # Multiple face detections for single frame given
        faces = np.tile(faces, (1, n_frames, 1)) # (n_faces, n_frames, 4)
      else:
        # Multiple face detections for multiple frames given
        if faces.shape[1] != n_frames:
          raise ValueError("Assuming detections of multiple faces for multiple frames given, but number of frames ({}) did not match number of detections for each face ({})".format(
            n_frames, faces.shape[1]))
  # Check that x0 < x1 and y0 < y1 for all faces
  if not (np.all((faces[...,2] - faces[...,0]) > 0) and np.all((faces[...,3] - faces[...,1]) > 0)):
    raise ValueError("Face detections are invalid, should be in form [x0, y0, x1, y1]")
  return faces

def check_faces_in_roi(
    faces: np.ndarray,
    roi: Union[np.ndarray, tuple, list],
    percentage_required_inside_roi: tuple = (0.5, 0.5)
  ) -> bool:
  """Check whether all faces are sufficiently inside the ROI.

  Args:
    faces: The faces. Shape (n_faces, 4) in form (x0, y0, x1, y1)
    roi: The region of interest. Shape (4,) in form (x0, y0, x1, y1)
    percentage_required_inside_roi: Tuple (w, h) indicating what percentage
      of width/height of face is required to remain inside the ROI.
  Returns:
    out: True if all faces are sufficiently inside the ROI.
  """
  faces_w = faces[:,2] - faces[:,0]
  faces_h = faces[:,3] - faces[:,1]
  faces_inside_roi = np.logical_and(
    np.logical_and(faces[:,2] - roi[0] > percentage_required_inside_roi[0] * faces_w,
                   roi[2] - faces[:,0] > percentage_required_inside_roi[0] * faces_w),
    np.logical_and(faces[:,3] - roi[1] > percentage_required_inside_roi[1] * faces_h,
                   roi[3] - faces[:,1] > percentage_required_inside_roi[1] * faces_h))
  facess_inside_roi = np.all(faces_inside_roi)
  return facess_inside_roi

def convert_ndarray_to_list(d: Union[dict, list, np.ndarray]):
  """Recursively convert any np.ndarray to list in nested object.
  
  Args:
    d: Nested object consisting of list, dict, and np.ndarray
  Returns:
    out: The same object with any np.ndarray converted to list
  """
  if isinstance(d, np.ndarray):
    return d.tolist()
  elif isinstance(d, dict):
    return {k: convert_ndarray_to_list(v) for k, v in d.items()}
  elif isinstance(d, list):
    return [convert_ndarray_to_list(i) for i in d]
  else:
    return d
  


# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime
from enum import IntEnum
import json
import logging
import numpy as np
import os
from prpy.constants import SECONDS_PER_MINUTE
from typing import Union

from vitallens.constants import DISCLAIMER
from vitallens.constants import CALC_HR_MIN, CALC_HR_MAX, CALC_HR_WINDOW_SIZE
from vitallens.constants import CALC_RR_MIN, CALC_RR_MAX, CALC_RR_WINDOW_SIZE
from vitallens.methods.g import GRPPGMethod
from vitallens.methods.chrom import CHROMRPPGMethod
from vitallens.methods.pos import POSRPPGMethod
from vitallens.methods.vitallens import VitalLensRPPGMethod
from vitallens.signal import windowed_freq, windowed_mean
from vitallens.ssd import FaceDetector
from vitallens.utils import load_config, probe_video_inputs, check_faces, convert_ndarray_to_list

class Method(IntEnum):
  VITALLENS = 1
  G = 2
  CHROM = 3
  POS = 4

logging.getLogger().setLevel("INFO")

class VitalLens:
  def __init__(
      self, 
      method: Method = Method.VITALLENS,
      api_key: str = None,
      detect_faces: bool = True,
      estimate_running_vitals: bool = True,
      fdet_max_faces: int = 1,
      fdet_fs: float = 1.0,
      fdet_score_threshold: float = 0.9,
      fdet_iou_threshold: float = 0.3,
      export_to_json: bool = True,
      export_dir: str = "."
    ):
    """Initialisation. Loads face detection model if necessary.

    Args:
      method: The rPPG method to be used for inference.
      api_key: Usage key for the VitalLens API (required for Method.VITALLENS)
      detect_faces: `True` if faces need to be detected, otherwise `False`.
      estimate_running_vitals: Set `True` to compute running vitals (e.g., `running_heart_rate`).
      fdet_max_faces: The maximum number of faces to detect (if necessary).
      fdet_fs: Frequency [Hz] at which faces should be scanned. Detections are
        linearly interpolated for remaining frames.
      fdet_score_threshold: Face detection score threshold.
      fdet_iou_threshold: Face detection iou threshold.
      export_to_json: If `True`, write results to a json file.
      export_dir: The directory to which json files are written.
    """
    self.api_key = api_key
    # Load the config and model
    self.config = load_config(method.name.lower() + ".yaml")
    self.method = method
    if self.config['model'] == 'g':
      self.rppg = GRPPGMethod(self.config)
    elif self.config['model'] == 'chrom':
      self.rppg = CHROMRPPGMethod(self.config)
    elif self.config['model'] == 'pos':
      self.rppg = POSRPPGMethod(self.config)
    elif self.config['model'] == 'vitallens':
      if self.api_key is None or self.api_key == '':
        raise ValueError("An API key is required to use Method.VITALLENS, but was not provided. "
                         "Get one for free at https://www.rouast.com/api.")
      self.rppg = VitalLensRPPGMethod(self.config, self.api_key)
    else:
      raise ValueError("Method {} not implemented!".format(self.config['model']))
    self.detect_faces = detect_faces
    self.estimate_running_vitals = estimate_running_vitals
    self.export_to_json = export_to_json
    self.export_dir = export_dir
    if detect_faces:
      # Initialise face detector
      self.face_detector = FaceDetector(
        max_faces=fdet_max_faces, fs=fdet_fs, score_threshold=fdet_score_threshold,
        iou_threshold=fdet_iou_threshold)
  def __call__(
      self,
      video: Union[np.ndarray, str],
      faces: Union[np.ndarray, list] = None,
      fps: float = None,
      override_fps_target: float = None,
      override_global_parse: bool = None,
      export_filename: str = None
    ) -> list:
    """Run rPPG inference.

    Args:
      video: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        with a sequence of frames in unscaled uint8 RGB format, or a path to a
        video file. Note that aggressive video encoding destroys the rPPG signal.
      faces: Face boxes in flat point form, containing [x0, y0, x1, y1] coords.
        Ignored unless detect_faces=False. Pass a list or np.ndarray of
        shape (n_faces, n_frames, 4) for multiple faces detected on multiple frames,
        shape (n_frames, 4) for single face detected on mulitple frames, or
        shape (4,) for a single face detected globally, or
        `None` to assume all frames already cropped to the same single face detection.
      fps: Sampling frequency of the input video. Required if type(video) == np.ndarray. 
      override_fps_target: Target fps at which rPPG inference should be run (optional).
        If not provided, will use default of the selected method.
      override_global_parse: If True, always use global parse. If False, don't use global parse.
        If None, choose based on video.
      export_filename: Filename for json export if applicable.
    Returns:
      result: Analysis results as a list of faces in the following format:
        [
          {
            'face': {
              'coordinates': <Face coordinates for each frame as np.ndarray of shape (n_frames, 4)>,
              'confidence': <Face live confidence for each frame as np.ndarray of shape (n_frames,)>,
              'note': <Explanatory note>
            },
            'vital_signs': {
              'heart_rate': {
                'value': <Estimated value as float scalar>,
                'unit': <Value unit>,
                'confidence': <Estimation confidence as float scalar>,
                'note': <Explanatory note>
              },
              'respiratory_rate': {
                'value': <Estimated value as float scalar>,
                'unit': <Value unit>,
                'confidence': <Estimation confidence as float scalar>,
                'note': <Explanatory note>
              },
              'ppg_waveform': {
                'data': <Estimated waveform value for each frame as np.ndarray of shape (n_frames,)>,
                'unit': <Data unit>,
                'confidence': <Estimation confidence for each frame as np.ndarray of shape (n_frames,)>,
                'note': <Explanatory note>
              },
              'respiratory_waveform': {
                'data': <Estimated waveform value for each frame as np.ndarray of shape (n_frames,)>,
                'unit': <Data unit>,
                'confidence': <Estimation confidence for each frame as np.ndarray of shape (n_frames,)>,
                'note': <Explanatory note>
              },
            },
            "message": <Message about estimates>
          },
          { 
            <same structure for face 2 if present>
          },
          ...
        ]
    """
    # Probe inputs
    inputs_shape, fps, _ = probe_video_inputs(video=video, fps=fps)
    # TODO: Optimize performance of simple rPPG methods for long videos
    # Warning if using long video
    target_fps = override_fps_target if override_fps_target is not None else self.rppg.fps_target
    if self.method != Method.VITALLENS and inputs_shape[0]/fps*target_fps > 3600:
      logging.warning("Inference for long videos has yet to be optimized for POS / G / CHROM. This may run out of memory and crash.")
    _, height, width, _ = inputs_shape
    if self.detect_faces:
      # Detect faces
      faces_rel, _ = self.face_detector(inputs=video, inputs_shape=inputs_shape, fps=fps)
      # If no faces detected: return empty list
      if len(faces_rel) == 0:
        logging.warning("No faces to analyze")
        return []
      # Convert to absolute units
      faces = (faces_rel * [width, height, width, height]).astype(np.int64)
      # Face axis first
      faces = np.transpose(faces, (1, 0, 2))
    # Check if the faces are valid
    faces = check_faces(faces, inputs_shape)  
    # Run separately for each face
    results = []
    for face in faces:
      # Run selected rPPG method
      data, unit, conf, note, live = self.rppg(
        frames=video, faces=face, fps=fps,
        override_fps_target=override_fps_target,
        override_global_parse=override_global_parse)
      # Parse face results
      face_result = {'face': {
        'coordinates': face,
        'confidence': live,
        'note': "Face detection coordinates for this face, along with live confidence levels."
      }}
      # Parse vital signs results
      vital_signs_results = {}
      for name in self.config['signals']:
        vital_signs_results[name] = {
          '{}'.format('data' if 'waveform' in name else 'value'): data[name],
          'unit': unit[name],
          'confidence': conf[name],
          'note': note[name]
        }
      if self.estimate_running_vitals:
        try:
          if 'ppg_waveform' in self.config['signals']:
            window_size = int(CALC_HR_WINDOW_SIZE*fps)
            running_hr = windowed_freq(
              x=data['ppg_waveform'], f_s=fps, f_res=0.005,
              f_range=(CALC_HR_MIN/SECONDS_PER_MINUTE, CALC_HR_MAX/SECONDS_PER_MINUTE),            
              window_size=window_size, overlap=window_size//2) * SECONDS_PER_MINUTE
            running_conf = windowed_mean(
              x=conf['ppg_waveform'], window_size=window_size, overlap=window_size//2)
            vital_signs_results['running_heart_rate'] = {
              'data': running_hr,
              'unit': 'bpm',
              'confidence': running_conf,
              'note': 'Estimate of the running heart rate using VitalLens, along with frame-wise confidences between 0 and 1.',
            }
          if 'respiratory_waveform' in self.config['signals']:
            window_size = int(CALC_RR_WINDOW_SIZE*fps)
            running_rr = windowed_freq(
              x=data['respiratory_waveform'], f_s=fps, f_res=0.005,
              f_range=(CALC_RR_MIN/SECONDS_PER_MINUTE, CALC_RR_MAX/SECONDS_PER_MINUTE),            
              window_size=window_size, overlap=window_size//2) * SECONDS_PER_MINUTE
            running_conf = windowed_mean(
              x=conf['respiratory_waveform'], window_size=window_size, overlap=window_size//2)
            vital_signs_results['running_respiratory_rate'] = {
              'data': running_rr,
              'unit': 'bpm',
              'confidence': running_conf,
              'note': 'Estimate of the running respiratory rate using VitalLens, along with frame-wise confidences between 0 and 1.',
            }
        except ValueError as e:
          logging.warning("Issue while computing running vitals: {}".format(e))
      face_result['vital_signs'] = vital_signs_results
      face_result['message'] = DISCLAIMER
      results.append(face_result)
    # Export to json
    if self.export_to_json:
      os.makedirs(self.export_dir, exist_ok=True)
      export_filename = "{}.json".format(export_filename) if export_filename is not None else 'vitallens_{}.json'.format(datetime.now().strftime("%Y%m%d_%H%M%S"))
      with open(os.path.join(self.export_dir, export_filename), 'w') as f:
        json.dump(convert_ndarray_to_list(results), f, indent=4)
    return results



# Copyright (c) 2024 Philipp Rouast
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import sys
sys.path.append('../vitallens-python')

from vitallens.methods.chrom import CHROMRPPGMethod
from vitallens.methods.g import GRPPGMethod
from vitallens.methods.pos import POSRPPGMethod
from vitallens.utils import load_config

@pytest.mark.parametrize("override_fps_target", [None, 15])
def test_CHROMRPPGMethod(request, override_fps_target):
  config = load_config("chrom.yaml")
  method = CHROMRPPGMethod(config)
  res = method.algorithm(np.random.rand(100, 3), fps=30.)
  assert res.shape == (100,)
  res = method.pulse_filter(res, fps=30.)
  assert res.shape == (100,)
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  data, unit, conf, note, live = method(
    frames=test_video_ndarray, faces=test_video_faces,
    fps=test_video_fps, override_fps_target=override_fps_target)
  assert all(key in data for key in method.signals)
  assert all(key in unit for key in method.signals)
  assert all(key in conf for key in method.signals)
  assert all(key in note for key in method.signals)
  assert data['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  assert conf['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  np.testing.assert_equal(conf['ppg_waveform'], np.ones((test_video_ndarray.shape[0],), np.float32))
  assert conf['heart_rate'] == 1.0
  np.testing.assert_equal(live, np.ones((test_video_ndarray.shape[0],), np.float32))

@pytest.mark.parametrize("override_fps_target", [None, 15])
def test_GRPPGMethod(request, override_fps_target):
  config = load_config("g.yaml")
  method = GRPPGMethod(config)
  res = method.algorithm(np.random.rand(100, 3), fps=30.)
  assert res.shape == (100,)
  res = method.pulse_filter(res, fps=30.)
  assert res.shape == (100,)
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  data, unit, conf, note, live = method(
    frames=test_video_ndarray, faces=test_video_faces,
    fps=test_video_fps, override_fps_target=override_fps_target)
  assert all(key in data for key in method.signals)
  assert all(key in unit for key in method.signals)
  assert all(key in conf for key in method.signals)
  assert all(key in note for key in method.signals)
  assert data['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  assert conf['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  np.testing.assert_equal(conf['ppg_waveform'], np.ones((test_video_ndarray.shape[0],), np.float32))
  assert conf['heart_rate'] == 1.0
  np.testing.assert_equal(live, np.ones((test_video_ndarray.shape[0],), np.float32))

@pytest.mark.parametrize("override_fps_target", [None, 15])
def test_POSRPPGMethod(request, override_fps_target):
  config = load_config("pos.yaml")
  method = POSRPPGMethod(config)
  res = method.algorithm(np.random.rand(100, 3), fps=30.)
  assert res.shape == (100,)
  res = method.pulse_filter(res, fps=30.)
  assert res.shape == (100,)
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  data, unit, conf, note, live = method(
    frames=test_video_ndarray, faces=test_video_faces,
    fps=test_video_fps, override_fps_target=override_fps_target)
  assert all(key in data for key in method.signals)
  assert all(key in unit for key in method.signals)
  assert all(key in conf for key in method.signals)
  assert all(key in note for key in method.signals)
  assert data['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  assert conf['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  np.testing.assert_equal(conf['ppg_waveform'], np.ones((test_video_ndarray.shape[0],), np.float32))
  assert conf['heart_rate'] == 1.0
  np.testing.assert_equal(live, np.ones((test_video_ndarray.shape[0],), np.float32))



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from prpy.constants import SECONDS_PER_MINUTE
from prpy.numpy.signal import moving_average_size_for_response, estimate_freq
from prpy.numpy.stride_tricks import window_view, resolve_1d_window_view
from typing import Tuple, Union

from vitallens.constants import CALC_HR_MAX, CALC_RR_MAX

def moving_average_size_for_hr_response(
    f_s: Union[float, int]
  ):
  return moving_average_size_for_response(f_s, CALC_HR_MAX / SECONDS_PER_MINUTE)

def moving_average_size_for_rr_response(
    f_s: Union[float, int]
  ):
  return moving_average_size_for_response(f_s, CALC_RR_MAX / SECONDS_PER_MINUTE)

def detrend_lambda_for_hr_response(
    f_s: Union[float, int]
  ):
  return int(0.1614*np.power(f_s, 1.9804))

def detrend_lambda_for_rr_response(
    f_s: Union[float, int]
  ):
  return int(4.4248*np.power(f_s, 2.1253))

def windowed_mean(
    x: np.ndarray,
    window_size: int,
    overlap: int
  ) -> np.ndarray:
  """Estimate the mean of an array using sliding windows. Returns same shape.
  
  Args:
    x: An array. Shape (n,)
    window_size: The size of the sliding window
    overlap: The overlap of subsequent locations of the sliding window
  Returns:
    out: The windowed mean. Shape (n,)
  """
  x = np.asarray(x)
  n = len(x)
  # Make sure there are enough vals
  if n <= window_size:
    raise ValueError("Not enough vals for frequency calculation.")
  else:
    # Generate a windowed view into x
    y, _, pad_end = window_view(
      x=x, min_window_size=window_size, max_window_size=window_size, overlap=overlap,
      pad_mode='reflect')
    # Estimate frequency for each window
    out = np.mean(y, axis=1)
    # Resolve to target dims
    out = resolve_1d_window_view(
      x=out, window_size=window_size, overlap=overlap, pad_end=pad_end, fill_method='start')
  # Make sure sizes match
  assert out.shape[0] == n, "out.shape[0] {} != {} n".format(
    out.shape[0], n)
  # Return
  return out

def windowed_freq(
    x: np.ndarray,
    window_size: int,
    overlap: int,
    f_s: Union[int, float],
    f_range: Tuple[Union[int, float], Union[int, float]] = None,
    f_res: Union[int, float] = None
  ) -> np.ndarray:
  """Estimate the varying frequency within a signal array using sliding windows. Returns same shape.
  
  Args:
    x: A signal with a frequency we want to estimate. Shape (n,)
    window_size: The size of the sliding window
    overlap: The overlap of subsequent locations of the sliding window
    f_s: The sampling frequency of x
    f_range: A range of (min, max) feasible frequencies to restrict the estimation to 
    f_res: The frequency resolution at which to estimate
  Returns:
    out: The estimated frequencies. Shape (n,)
  """
  x = np.asarray(x)
  n = len(x)
  # Make sure there are enough vals
  if n <= window_size:
    raise ValueError("Not enough vals for frequency calculation.")
  else:
    # Generate a windowed view into x
    y, _, pad_end = window_view(
      x=x, min_window_size=window_size, max_window_size=window_size, overlap=overlap,
      pad_mode='reflect')
    # Estimate frequency for each window
    freqs = estimate_freq(
      y, f_s=f_s, f_range=f_range, f_res=f_res, method='periodogram', axis=1)
    # Resolve to target dims
    freq_vals = resolve_1d_window_view(
      x=freqs, window_size=window_size, overlap=overlap, pad_end=pad_end, fill_method='start')
  # Make sure sizes match
  assert freq_vals.shape[0] == n, "freq_vals.shape[0] {} != {} n".format(
    freq_vals.shape[0], n)
  # Return
  return freq_vals

def reassemble_from_windows(
    x: np.ndarray,
    idxs: np.ndarray
  ) -> np.ndarray:
  """Reassemble windowed data using corresponding idxs.

  Args:
    x: Data generated using a windowing operation. Shape (n_windows, n, window_size)
    idxs: Indices of x in the original 1-d array. Shape (n_windows, window_size)
  
  Returns:
    out: Reassembled data. Shape (n, n_idxs)
    idxs: Reassembled idxs. Shape (n_idxs)
  """
  x = np.asarray(x)
  idxs = np.asarray(idxs)
  # Transpose x (n, n_windows, window_size)
  x = np.transpose(x, (1, 0, 2))
  # Adjust indices based on their window position
  offset_idxs = idxs - np.arange(idxs.shape[0])[:, np.newaxis]
  # Find strictly increasing indices using np.maximum.accumulate
  flat_offset_idxs = offset_idxs.flatten()
  max_so_far = np.maximum.accumulate(flat_offset_idxs.flatten())
  mask = (flat_offset_idxs == max_so_far)  # Mask to keep only strictly increasing indices
  # Filter data based on mask and extract the final result values
  result = x.reshape(x.shape[0], -1)[:,mask]
  idxs = idxs.flatten()[mask]
  return result, idxs



# Copyright (c) 2024 Philipp Rouast
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
import json
import numpy as np
import pytest
import requests
from unittest.mock import Mock, patch

import sys
sys.path.append('../vitallens-python')

from vitallens.constants import API_MAX_FRAMES, API_MIN_FRAMES, API_URL
from vitallens.methods.vitallens import VitalLensRPPGMethod
from vitallens.utils import load_config, parse_video_inputs

def create_mock_response(
    status_code: int,
    json_data: dict
  ) -> Mock:
  """Create a mock response
  
  Args:
    status_code: The desired status code
    json_data: The desired json data
  Returns:
    mock: The Mock response object
  """
  mock_resp = Mock()
  mock_resp.status_code = status_code
  mock_resp.text = json.dumps(json_data)
  mock_resp.json.return_value = json_data
  return mock_resp

def create_mock_api_response(
    url: str,
    headers: dict,
    json: dict
  ) -> Mock:
  """Create a mock api response
  
  Args:
    url: The API URL
    headers: The request headers
    json: The request payload
  Returns:
    mock: The Mock response object
  """
  api_key = headers["x-api-key"]
  if api_key is None or not isinstance(api_key, str) or len(api_key) < 30:
    return create_mock_response(
      status_code=403, json_data={"vital_signs": None, "face": None, "state": None, "message": "Error"}) 
  video_base64 = json["video"]
  if video_base64 is None:
    return create_mock_response(
      status_code=400, json_data={"vital_signs": None, "face": None, "state": None, "message": "Error"})
  try:
    video = np.frombuffer(base64.b64decode(video_base64), dtype=np.uint8)
    video = video.reshape((-1, 40, 40, 3))
  except Exception as e:
    return create_mock_response(status_code=400, json_data={"vital_signs": None, "face": None, "state": None, "message": "Error: {}".format(e)})
  if video.shape[0] < API_MIN_FRAMES or video.shape[0] > API_MAX_FRAMES:
    return create_mock_response(status_code=400, json_data={"vital_signs": None, "face": None, "state": None, "message": "Error"})
  return create_mock_response(
      status_code=200,
      json_data={
        "vital_signs": {
          "heart_rate": {"value": 60.0, "unit": "bpm", "confidence": 0.99, "note": "Note"},
          "respiratory_rate": {"value": 15.0, "unit": "bpm", "confidence": 0.97, "note": "Note"},
          "ppg_waveform": {"data": np.random.rand(video.shape[0]).tolist(), "unit": "unitless", "confidence": np.ones(video.shape[0]).tolist(), "note": "Note"},
          "respiratory_waveform": {"data": np.random.rand(video.shape[0]).tolist(), "unit": "unitless", "confidence": np.ones(video.shape[0]).tolist(), "note": "Note"}},
        "face": {"confidence": np.random.rand(video.shape[0]).tolist(), "note": "Note"},
        "state": {"data": np.zeros((2, 128), dtype=np.float32).tolist(), "note": "Note"},
        "message": "Message"})

@pytest.mark.parametrize("file", [True, False])
@pytest.mark.parametrize("override_fps_target", [None, 15, 10])
@pytest.mark.parametrize("long", [False, True])
@pytest.mark.parametrize("override_global_parse", [False, True])
@patch('requests.post', side_effect=create_mock_api_response)
def test_VitalLensRPPGMethod_mock(mock_post, request, file, long, override_fps_target, override_global_parse):
  if long and file:
    pytest.skip("Skip because parameter combination does not work")
  config = load_config("vitallens.yaml")
  api_key = request.getfixturevalue('test_dev_api_key')
  test_video_path = request.getfixturevalue('test_video_path')
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  method = VitalLensRPPGMethod(config, api_key=api_key)
  if file:
    data, unit, conf, note, live = method(
      frames=test_video_path, faces=test_video_faces,
      override_fps_target=override_fps_target,
      override_global_parse=override_global_parse)
  else:
    if long:
      n_repeats = (API_MAX_FRAMES * 3) // test_video_ndarray.shape[0] + 1
      test_video_ndarray = np.repeat(test_video_ndarray, repeats=n_repeats, axis=0)
      test_video_faces = np.repeat(test_video_faces, repeats=n_repeats, axis=0)
    data, unit, conf, note, live = method(
      frames=test_video_ndarray, faces=test_video_faces,
      fps=test_video_fps, override_fps_target=override_fps_target,
      override_global_parse=override_global_parse)
  assert all(key in data for key in method.signals)
  assert all(key in unit for key in method.signals)
  assert all(key in conf for key in method.signals)
  assert all(key in note for key in method.signals)
  assert data['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  assert conf['ppg_waveform'].shape == (test_video_ndarray.shape[0],)
  assert data['respiratory_waveform'].shape == (test_video_ndarray.shape[0],)
  assert conf['respiratory_waveform'].shape == (test_video_ndarray.shape[0],)
  assert live.shape == (test_video_ndarray.shape[0],)

@pytest.mark.parametrize("process_signals", [True, False])
def test_VitalLens_API_valid_response(request, process_signals):
  config = load_config("vitallens.yaml")
  api_key = request.getfixturevalue('test_dev_api_key')
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  frames, *_ = parse_video_inputs(
    video=test_video_ndarray, fps=test_video_fps, target_size=config['input_size'],
    roi=test_video_faces[0].tolist(), library='prpy', scale_algorithm='bilinear')
  headers = {"x-api-key": api_key}
  payload = {"video": base64.b64encode(frames[:16].tobytes()).decode('utf-8')}
  if process_signals: payload['fps'] = str(30)
  response = requests.post(API_URL, headers=headers, json=payload)
  response_body = json.loads(response.text)
  assert response.status_code == 200
  assert all(key in response_body for key in ["face", "vital_signs", "state", "message"])
  vital_signs = response_body["vital_signs"]
  assert all(key in vital_signs for key in ["ppg_waveform", "respiratory_waveform"])
  ppg_waveform_data = np.asarray(response_body["vital_signs"]["ppg_waveform"]["data"])
  ppg_waveform_conf = np.asarray(response_body["vital_signs"]["ppg_waveform"]["confidence"])
  resp_waveform_data = np.asarray(response_body["vital_signs"]["respiratory_waveform"]["data"])
  resp_waveform_conf = np.asarray(response_body["vital_signs"]["respiratory_waveform"]["confidence"])
  assert ppg_waveform_data.shape == (16,)
  assert ppg_waveform_conf.shape == (16,)
  assert resp_waveform_data.shape == (16,)
  assert resp_waveform_conf.shape == (16,)
  assert all((key in vital_signs) if process_signals else (key not in vital_signs) for key in ["heart_rate", "respiratory_rate"])
  live = np.asarray(response_body["face"]["confidence"])
  assert live.shape == (16,)
  state = np.asarray(response_body["state"]["data"])
  assert state.shape == (2, 128)

def test_VitalLens_API_wrong_api_key(request):
  config = load_config("vitallens.yaml")
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  test_video_fps = request.getfixturevalue('test_video_fps')
  test_video_faces = request.getfixturevalue('test_video_faces')
  frames, *_ = parse_video_inputs(
    video=test_video_ndarray, fps=test_video_fps, target_size=config['input_size'],
    roi=test_video_faces[0].tolist(), library='prpy', scale_algorithm='bilinear')
  headers = {"x-api-key": "WRONG_API_KEY"}
  payload = {"video": base64.b64encode(frames[:16].tobytes()).decode('utf-8')}
  response = requests.post(API_URL, headers=headers, json=payload)
  assert response.status_code == 403

def test_VitalLens_API_no_video(request):
  api_key = request.getfixturevalue('test_dev_api_key')
  headers = {"x-api-key": api_key}
  payload = {"some_key": "irrelevant"}
  response = requests.post(API_URL, headers=headers, json=payload)
  assert response.status_code == 400

def test_VitalLens_API_no_parseable_video(request):
  api_key = request.getfixturevalue('test_dev_api_key')
  headers = {"x-api-key": api_key}
  payload = {"video": "not_parseable"}
  response = requests.post(API_URL, headers=headers, json=payload)
  assert response.status_code == 400
